# decode.py

"""
Krock32 Decoder

Class definition for the Krock32 decoder, taking a Crockford Base32 string
and turning it into bytes-data.
"""

from collections import namedtuple


class DecoderAlreadyFinalizedException(Exception):
    pass


class DecoderInvalidStringLengthException(Exception):
    pass


class DecoderNonZeroCarryException(Exception):
    pass


class DecoderChecksumException(Exception):
    pass


class Decoder:
    def __init__(
        self,
        strict: bool = False,
        ignore_non_alphabet: bool = True,
        checksum: bool = False,
    ):
        self._string_buffer: str = ""
        self._bytearray: bytearray = bytearray()
        self._strict: bool = strict
        self._ignore_non_alphabet: bool = ignore_non_alphabet
        self._alphabet = self._make_alphabet(
            "0123456789ABCDEFGHJKMNPQRSTVWXYZ*~$=U", strict=self._strict
        )
        self._is_finished: bool = False
        self._p_byte = namedtuple("ProcessedByte", ["byte", "carry"])
        self._do_checksum: bool = checksum
        self._checksum: int = 0

    def _make_alphabet(self, alphabet_string: str, strict: bool) -> dict:
        alphabet = {}
        for i, x in enumerate(alphabet_string):
            alphabet[x.upper()] = i
            if not strict:
                alphabet[x.lower()] = i
        if not strict:
            alphabet["O"], alphabet["o"] = 0, 0
            (alphabet["I"], alphabet["i"], alphabet["L"], alphabet["l"]) = 1, 1, 1, 1
        return alphabet

    def _update_checksum(self, byte: int):
        if not self._do_checksum:
            return
        self._checksum = ((self._checksum << 8) + byte) % 37

    def _decode_first_byte(self, symbols: str) -> tuple:
        byte = self._alphabet.get(symbols[0]) << 3
        second_sym = self._alphabet.get(symbols[1])
        byte += second_sym >> 2
        carry = second_sym & 0b11
        self._update_checksum(byte)
        return self._p_byte(byte=byte, carry=carry)

    def _decode_second_byte(self, symbols: str, carry: int) -> tuple:
        byte = (self._alphabet.get(symbols[0]) << 1) + (carry << 6)
        second_sym = self._alphabet.get(symbols[1])
        byte += second_sym >> 4
        carry = second_sym & 0b1111
        self._update_checksum(byte)
        return self._p_byte(byte=byte, carry=carry)

    def _decode_third_byte(self, symbols: str, carry: int) -> tuple:
        sym = self._alphabet.get(symbols)
        byte = (sym >> 1) + (carry << 4)
        carry = sym & 1
        self._update_checksum(byte)
        return self._p_byte(byte=byte, carry=carry)

    def _decode_fourth_byte(self, symbols: str, carry: int) -> tuple:
        byte = (self._alphabet.get(symbols[0]) << 2) + (carry << 7)
        second_sym = self._alphabet.get(symbols[1])
        byte += second_sym >> 3
        carry = second_sym & 0b111
        self._update_checksum(byte)
        return self._p_byte(byte=byte, carry=carry)

    def _decode_fifth_byte(self, symbols: str, carry: int) -> tuple:
        byte = (carry << 5) + self._alphabet.get(symbols)
        self._update_checksum(byte)
        return self._p_byte(byte=byte, carry=0)

    def _return_quantum(
        self, quantum: str, p_byte: tuple, array: bytearray
    ) -> bytearray:
        if p_byte.carry == 0:
            return array
        else:
            raise DecoderNonZeroCarryException(
                "Quantum %s decoded with non-zero carry %i" % (quantum, p_byte.carry)
            )

    def _decode_quantum(self, quantum: str) -> bytearray:
        if not len(quantum) in [2, 4, 5, 7, 8]:
            raise DecoderInvalidStringLengthException
        buffer = bytearray()
        p_byte = self._decode_first_byte(quantum[0:2])
        buffer.append(p_byte.byte)
        if len(quantum) == 2:
            return self._return_quantum(quantum, p_byte, buffer)
        p_byte = self._decode_second_byte(quantum[2:4], p_byte.carry)
        buffer.append(p_byte.byte)
        if len(quantum) == 4:
            return self._return_quantum(quantum, p_byte, buffer)
        p_byte = self._decode_third_byte(quantum[4], p_byte.carry)
        buffer.append(p_byte.byte)
        if len(quantum) == 5:
            return self._return_quantum(quantum, p_byte, buffer)
        p_byte = self._decode_fourth_byte(quantum[5:7], p_byte.carry)
        buffer.append(p_byte.byte)
        if len(quantum) == 7:
            return self._return_quantum(quantum, p_byte, buffer)
        p_byte = self._decode_fifth_byte(quantum[7], p_byte.carry)
        buffer.append(p_byte.byte)
        return buffer

    def _consume(self):
        tail = 0
        for head in range(8, len(self._string_buffer), 8):
            quantum: str = self._string_buffer[tail:head]
            self._bytearray.extend(self._decode_quantum(quantum))
            tail = head
        self._string_buffer = self._string_buffer[tail:]

    def update(self, string: str):
        if self._is_finished:
            raise DecoderAlreadyFinalizedException
        self._string_buffer += string
        self._consume()

    def _check_checksum(self, checksymbol: str) -> bytes:
        expected = self._alphabet.get(checksymbol)
        if self._checksum == expected:
            return bytes(self._bytearray)
        else:
            raise DecoderChecksumException(
                "Calculated checksum %i, expected %i" % (self._checksum, expected)
            )

    def finalize(self) -> bytes:
        if self._is_finished:
            raise DecoderAlreadyFinalizedException
        self._is_finished = True
        if self._do_checksum:
            checksymbol = self._string_buffer[-1]
            self._string_buffer = self._string_buffer[:-1]
        self._bytearray.extend(
            self._decode_quantum(self._string_buffer)
            if len(self._string_buffer) > 0
            else []
        )
        if self._do_checksum:
            return self._check_checksum(checksymbol)
        else:
            return bytes(self._bytearray)



# __init__.py

"""
Krock32 module 0.1.1

Implementation of Crockford's base32 alphabet encoding, decoding,
and pretty printing
"""

# pylama:ignore=W0611
from krock32.decode import Decoder
from krock32.encode import Encoder

__version__ = "0.1.1"



# encode.py

"""
Krock32 encoder

Class definition for the krock32 encoder, taking bytes-data and turning it
into Crockford's Base32 string.
"""

from collections import namedtuple
from typing import List


class EncoderAlreadyFinalizedException(Exception):
    pass


class Encoder:
    def __init__(self, checksum: bool = False):
        self._string_list: List[str] = []
        self._byte_buffer: bytearray = bytearray()
        self._alphabet: dict = self._make_alphabet(
            "0123456789ABCDEFGHJKMNPQRSTVWXYZ*~$=U"
        )
        self._is_finished: bool = False
        self._checksum = 0
        self._do_checksum: bool = checksum
        self._p_quin = namedtuple("ProcessedQuin", ["sym", "rem"])

    def _make_alphabet(self, alphabet_string: str) -> dict:
        alphabet = {}
        for i, x in enumerate(alphabet_string):
            alphabet[i] = x
        return alphabet

    def _update_checksum(self, byte: int):
        if not self._do_checksum:
            return
        self._checksum = ((self._checksum << 8) + byte) % 37

    def _encode_first_quin(self, byte) -> tuple:
        self._update_checksum(byte)
        quin = byte >> 3
        rem = (byte & 0b111) << 2
        return self._p_quin(sym=self._alphabet.get(quin), rem=rem)

    def _encode_second_quin(self, byte, remainder) -> tuple:
        self._update_checksum(byte)
        sym = ""
        quin = (byte >> 6) + remainder
        sym += self._alphabet.get(quin)
        quin = (byte >> 1) & 0b11111
        rem = (byte & 0b1) << 4
        sym += self._alphabet.get(quin)
        return self._p_quin(sym=sym, rem=rem)

    def _encode_third_quin(self, byte, remainder) -> tuple:
        self._update_checksum(byte)
        quin = (byte >> 4) + remainder
        rem = (byte & 0b1111) << 1
        return self._p_quin(sym=self._alphabet.get(quin), rem=rem)

    def _encode_fourth_quin(self, byte, remainder) -> tuple:
        self._update_checksum(byte)
        sym = ""
        quin = (byte >> 7) + remainder
        sym += self._alphabet.get(quin)
        quin = (byte >> 2) & 0b11111
        sym += self._alphabet.get(quin)
        rem = (byte & 0b11) << 3
        return self._p_quin(sym=sym, rem=rem)

    def _encode_fifth_quin(self, byte, remainder) -> tuple:
        self._update_checksum(byte)
        sym = ""
        quin = (byte >> 5) + remainder
        sym += self._alphabet.get(quin)
        quin = byte & 0b11111
        sym += self._alphabet.get(quin)
        return self._p_quin(sym=sym, rem=0)

    def _encode_quantum(self, quantum: bytearray) -> List[str]:
        slist: List[str] = []
        p_quin = self._encode_first_quin(quantum[0])
        slist.append(p_quin.sym)
        if len(quantum) == 1:
            slist.append(self._alphabet.get(p_quin.rem))
            return slist
        p_quin = self._encode_second_quin(quantum[1], p_quin.rem)
        slist.append(p_quin.sym)
        if len(quantum) == 2:
            slist.append(self._alphabet.get(p_quin.rem))
            return slist
        p_quin = self._encode_third_quin(quantum[2], p_quin.rem)
        slist.append(p_quin.sym)
        if len(quantum) == 3:
            slist.append(self._alphabet.get(p_quin.rem))
            return slist
        p_quin = self._encode_fourth_quin(quantum[3], p_quin.rem)
        slist.append(p_quin.sym)
        if len(quantum) == 4:
            slist.append(self._alphabet.get(p_quin.rem))
            return slist
        p_quin = self._encode_fifth_quin(quantum[4], p_quin.rem)
        slist.append(p_quin.sym)
        return slist

    def _consume(self):
        tail = 0
        for head in range(5, len(self._byte_buffer), 5):
            quantum = self._byte_buffer[tail:head]
            self._string_list.extend(self._encode_quantum(quantum))
            tail = head
        self._byte_buffer = self._byte_buffer[tail:]

    def update(self, data: bytes):
        if self._is_finished:
            raise EncoderAlreadyFinalizedException
        self._byte_buffer.extend(bytearray(data))
        self._consume()

    def finalize(self) -> str:
        if self._is_finished:
            raise EncoderAlreadyFinalizedException
        self._is_finished = True
        self._string_list.extend(
            self._encode_quantum(self._byte_buffer)
            if len(self._byte_buffer) > 0
            else []
        )
        if self._do_checksum:
            self._string_list.append(self._alphabet.get(self._checksum))
        return "".join(self._string_list)



import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

setup(
    name='krock32',
    version='0.1.1',
    description='Base32 encoder/decoder using Crockford\'s alphabet',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/DrSLDR/krock32',
    author='Jonas A. Hultén',
    author_email='sldr@sldr.se',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8'
    ],
    packages=['krock32'],
    include_package_data=True
)
